from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

import h5py
import s3fs
from cpdd_dataset.intrinsics import CylindricalIntrinsics, SphericalIntrinsics, PinholeIntrinsics, Intrinsics
from ._side import Side


class DataSource(ABC):
    @property
    @abstractmethod
    def data(self) -> Data:
        pass

    def __enter__(self) -> Data:
        if not self.is_downloaded:
            raise ValueError(f"{self} is  not downloaded")

        if hasattr(self, "_open_data"):
            return self._open_data

        self._open_data = self.data
        return self._open_data

    def __exit__(self, exc_type, exc_val: Data, exc_tb):
        self._open_data.close()
        del self._open_data

    @abstractmethod
    def download(self, force: bool = False) -> DataSource:
        pass

    @abstractmethod
    def is_downloaded(self) -> bool:
        pass


class PoseData:
    def __init__(self, file: h5py.File):
        self._file = file

    def close(self):
        return self._file.close()

    @property
    def absolute_pose(self) -> h5py.Dataset:
        """
        The absolute pose (origin is determined by the simulator, and is not significant)

        :return: A (frames, 6) float32 Dataset
        """
        return self._file["abs_pose"]

    @property
    def relative_pose(self) -> h5py.Dataset:
        """
        The pose relative to the previous pose

        :return: A (frames, 6) float32 Dataset
        """
        return self._file["rel_pose"]

    @property
    def start_relative_pose(self) -> h5py.Dataset:
        """
        The pose relative to the initial pose

        :return: A (frames, 6) float32 Dataset
        """
        return self._file["start_rel_pose"]


class Data:
    def __init__(self, file: h5py.File, data: h5py.Group, intrinsics: Intrinsics):
        self._data: h5py.Group = data
        self._file: h5py.File = file
        self._intrinsics = intrinsics

    @property
    def color(self) -> h5py.Dataset:
        """
        :return: A (frames, height, width, 3) uint8 Dataset
        """
        return self._data["rgb"]

    @property
    def depth(self) -> h5py.Dataset:
        """
        Depth is measured in dm (10th of a meter).
        :return: A (frames, height, width, 1) uint16 Dataset
        """
        return self._data["depth"]

    @property
    def intrinsics(self) -> Intrinsics:
        return self._intrinsics

    def close(self):
        return self._file.close()


class SplitData:
    def __init__(self, file: h5py.File, intrinsics: Intrinsics):
        self._intrinsics = intrinsics
        self._file: h5py.File = file

    @property
    def top(self) -> Data:
        return Data(self._file, self._file["top"], self._intrinsics)

    @property
    def bottom(self) -> Data:
        return Data(self._file, self._file["bottom"], self._intrinsics)

    @property
    def left(self) -> Data:
        return Data(self._file, self._file["left"], self._intrinsics)

    @property
    def right(self) -> Data:
        return Data(self._file, self._file["right"], self._intrinsics)

    @property
    def front(self) -> Data:
        return Data(self._file, self._file["front"], self._intrinsics)

    @property
    def back(self) -> Data:
        return Data(self._file, self._file["back"], self._intrinsics)

    def __getitem__(self, item: Side):
        return Data(self._file, self._file[item.name.lower()], self._intrinsics)


class DataFile(ABC):
    def __init__(self, config):
        self._config = config

    @property
    @abstractmethod
    def filename(self) -> str:
        pass

    @property
    def download_file(self) -> Path:
        return self._config.download_location / self.filename

    @property
    def download_file_if_exists(self) -> Path:
        if not self.is_downloaded:
            raise ValueError(f"{self} is  not downloaded")

        return self._config.download_location / self.filename

    @property
    def remote_location(self) -> str:
        return self._config.remote_location + self.filename

    @property
    def is_downloaded(self) -> bool:
        return self.download_file.exists()

    @abstractmethod
    def download(self, force: bool = False) -> DataFile:
        pass

    @property
    @abstractmethod
    def intrinsics(self) -> Intrinsics:
        pass

    @property
    def remote_exists(self) -> bool:
        if self.is_downloaded:
            return True

        fs = s3fs.S3FileSystem()

        return fs.exists(self.remote_location)


    def _download(self, force: bool = False):
        if self.is_downloaded and not force:
            return

        if self.is_downloaded and self.download_file.exists():
            self.download_file.unlink()

        self.download_file.parent.mkdir(parents=True, exist_ok=True)

        fs = s3fs.S3FileSystem()
        fs.get(self.remote_location, str(self.download_file))


class PoseDataFile(DataFile):

    @property
    def intrinsics(self) -> Intrinsics:
        return None

    @property
    def filename(self) -> str:
        return "pose.hdf5"

    def download(self, force: bool = False) -> DataFile:
        self._download(force)
        return self

    @property
    def data(self) -> PoseData:
        return PoseData(h5py.File(self.download_file_if_exists, 'r'))

    def __enter__(self) -> PoseData:
        if not self.is_downloaded:
            raise ValueError(f"{self} is  not downloaded")

        if hasattr(self, "_open_data"):
            return self._open_data

        self._open_data = self.data
        return self._open_data

    def __exit__(self, exc_type, exc_val: Data, exc_tb):
        self._open_data.close()
        del self._open_data


class CylindricalDataFile(DataFile, DataSource):

    def __init__(self, config):
        super().__init__(config)

    def __repr__(self):
        return f"CylindricalDataFile(config={self._config}, location={str(self.download_file)}, downloaded=" \
               f"{self.is_downloaded})"

    @property
    def filename(self) -> str:
        return "cylindrical.hdf5"

    @property
    def data(self) -> Data:
        file = h5py.File(self.download_file_if_exists, 'r')
        return Data(file, file, self.intrinsics)

    @property
    def intrinsics(self) -> CylindricalIntrinsics:
        return CylindricalIntrinsics()

    def download(self, force: bool = False) -> CylindricalDataFile:
        self._download(force)
        return self


class SphericalDataFile(DataFile, DataSource):
    def __init__(self, config):
        super().__init__(config)

    def __repr__(self):
        return f"SphericalDataFile(config={self._config}, location={str(self.download_file)}, down" \
               f"loaded={self.is_downloaded})"

    @property
    def filename(self) -> str:
        return "spherical.hdf5"

    @property
    def data(self) -> Data:
        file = h5py.File(self.download_file_if_exists, 'r')
        return Data(file, file, self.intrinsics)

    @property
    def intrinsics(self) -> SphericalIntrinsics:
        return SphericalIntrinsics()

    def download(self, force: bool = False) -> SphericalDataFile:
        self._download(force)
        return self


class PinholeDataFile(DataFile):

    def __init__(self, config):
        super().__init__(config)

    def __repr__(self):
        return f"PinholeDataFile(config={self._config}, location={str(self.download_file)}, down" \
               f"loaded={self.is_downloaded})"

    @property
    def filename(self) -> str:
        return "pinhole.hdf5"

    def __enter__(self) -> SplitData:
        if not self.is_downloaded:
            raise ValueError(f"{self} is  not downloaded")

        file = h5py.File(self.download_file_if_exists, 'r')
        self._open_file = file
        return SplitData(file, self.intrinsics)

    def __exit__(self, exc_type, exc_val: SplitData, exc_tb):
        self._open_file.close()

    @property
    def intrinsics(self) -> PinholeIntrinsics:
        return PinholeIntrinsics()

    @property
    def top(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "top")

    @property
    def bottom(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "bottom")

    @property
    def left(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "left")

    @property
    def right(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "right")

    @property
    def front(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "front")

    @property
    def back(self) -> PinholeDataFileSide:
        return PinholeDataFileSide(self, "back")

    def __getitem__(self, item: Side):
        return PinholeDataFileSide(self, item.name.lower())

    def download(self, force: bool = False) -> PinholeDataFile:
        self._download(force)
        return self


@dataclass
class PinholeDataFileSide(DataSource):
    data_file: PinholeDataFile
    side: str

    @property
    def intrinsics(self) -> PinholeIntrinsics:
        return self.data_file.intrinsics

    @property
    def data(self) -> Data:
        file = h5py.File(self.data_file.download_file_if_exists, 'r')
        return Data(file, file[self.side], self.data_file.intrinsics)

    def download(self, force: bool = False) -> PinholeDataFileSide:
        self.data_file.download(force)
        return self

    def is_downloaded(self) -> bool:
        return self.data_file.is_downloaded



from ._location import get_download_location, set_download_location
from ._run_data import CylindricalDataFile, Data, DataFile, DataSource, PinholeDataFile, PinholeDataFileSide, \
    SphericalDataFile, SplitData
from ._side import Side
from ._crop import crop_pinhole_to_90



import cv2
import numpy as np

from cpdd_dataset.intrinsics import PinholeIntrinsics

_new_size = 2 * PinholeIntrinsics().f_x * np.tan(np.pi / 4)
_scale_factor = 768 / _new_size


def crop_pinhole_to_90(image):
    image = cv2.resize(image, dsize=(0, 0), fx=_scale_factor, fy=_scale_factor)
    center = tuple(p // 2 for p in image.shape[:2])

    image = cv2.getRectSubPix(image, (768, 768), center)
    return image



from pathlib import Path
from typing import Union
import os


def get_download_location() -> Path:
    if os.getenv("CARLA_DOWNLOAD_LOCATION") is None:
        set_download_location("~/.cpdd_dataset")

    return Path(os.getenv("CARLA_DOWNLOAD_LOCATION")).expanduser()


def set_download_location(new_loc: Union[str, Path]):
    if not isinstance(new_loc, Path):
        new_loc = Path(new_loc)

    new_loc.mkdir(parents=True, exist_ok=True)
    os.environ["CARLA_DOWNLOAD_LOCATION"] = str(new_loc)



from abc import ABC, abstractmethod
import numpy as np


class Intrinsics(ABC):

    @property
    @abstractmethod
    def f_x(self) -> np.float32:
        pass

    @property
    @abstractmethod
    def f_y(self) -> np.float32:
        pass

    @property
    @abstractmethod
    def c_x(self) -> np.float32:
        pass

    @property
    @abstractmethod
    def c_y(self) -> np.float32:
        pass

    @property
    @abstractmethod
    def height(self) -> int:
        pass

    @property
    @abstractmethod
    def width(self) -> int:
        pass

    @property
    def K(self) -> np.ndarray:
        return np.array([[self.f_x, 0, self.c_x, 0],
                         [0, self.f_y, self.c_y, 0],
                         [0, 0, 1, 0],
                         [0, 0, 0, 1]], dtype='float32')

    @property
    def normalized_K(self):
        K = self.K
        K[0, :] /= self.width
        K[1, :] /= self.height
        return K



class CylindricalIntrinsics(Intrinsics):
    @property
    def height(self) -> int:
        return 1024

    @property
    def width(self) -> int:
        return 2048

    @property
    def f_x(self) -> np.float32:
        return np.float32(325.949323452201668)

    @property
    def c_x(self) -> np.float32:
        return np.float32(1024.000000000000000)

    @property
    def f_y(self) -> np.float32:
        return np.float32(1023.000000000000000)

    @property
    def c_y(self) -> np.float32:
        return np.float32(511.500000000000000)


class SphericalIntrinsics(Intrinsics):
    @property
    def height(self) -> int:
        return 1024

    @property
    def width(self) -> int:
        return 2048

    @property
    def f_x(self) -> np.float32:
        return np.float32(325.949323452201668)

    @property
    def c_x(self) -> np.float32:
        return np.float32(1024.000000000000000)

    @property
    def f_y(self) -> np.float32:
        return np.float32(325.949323452201668)

    @property
    def c_y(self) -> np.float32:
        return np.float32(512.000000000000000)


class PinholeIntrinsics(Intrinsics):
    @property
    def f_x(self) -> np.float32:
        return np.float32(322.2142583720755)

    @property
    def f_y(self) -> np.float32:
        return np.float32(322.2142583720755)

    @property
    def c_x(self) -> np.float32:
        return np.float32(384.0)

    @property
    def c_y(self) -> np.float32:
        return np.float32(384.0)

    @property
    def height(self) -> int:
        return 768

    @property
    def width(self) -> int:
        return 768

    @property
    def fov(self) -> int:
        return 100


class Pinhole90Intrinsics(Intrinsics):
    @property
    def f_x(self) -> np.float32:
        return np.float32(384.0)

    @property
    def f_y(self) -> np.float32:
        return np.float32(384.0)

    @property
    def c_x(self) -> np.float32:
        return np.float32(384.0)

    @property
    def c_y(self) -> np.float32:
        return np.float32(384.0)

    @property
    def height(self) -> int:
        return 768

    @property
    def width(self) -> int:
        return 768

    @property
    def fov(self) -> int:
        return 90



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpdd_dataset',
 'cpdd_dataset.config',
 'cpdd_dataset.data',
 'cpdd_dataset.intrinsics']

package_data = \
{'': ['*']}

install_requires = \
['h5py>=2.10.0,<3.0.0',
 'numpy>=1.18.1,<2.0.0',
 'pandas>=1.0.0,<2.0.0',
 's3fs>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'cpdd-dataset',
    'version': '0.1.3',
    'description': 'Accessors for CARLA Panoramic Depth Detection Dataset, the dataset created as part of my thesis',
    'long_description': '\n# CPDD-Dataset\n\nThis is the data access package for the CARLA panoramaic depth detection dataset,\ncreated and published as part of my thesis *Dataset and Evaluation of Self-Supervised Learning for Panoramic Depth Estimation*\n\nThe code used to generate the dataset is available at https://github.com/rnett/CARLASim.\n\n## Example\n\nSet the environment variable `CARLA_DOWNLOAD_LOCATION` to wherever you want to keep the data.\n\nThen, loading the data is as simple as:\n\n```python\nimport cpdd_dataset\n\ntrain_configs = cpdd_dataset.config.load_csv("train_data.csv")\ntrain_data = []\n\nfor config in train_configs:\n    with config.cylindrical_data.download() as data:\n        train_data.append({"color": data.color, "depth": data.depth})\n```\n\nYou can make a csv file structured like:\n```csv\ncity,rain,time,num_cars,num_peds,index\nTown01,Clear,Noon,*,*,2\n```\nand load it in a single line: `carla_dataset.data.load_csv("train.csv")`.\n\nIf you need camera intrinsics, `CylindricalIntrinsics`, `SphericalIntrinsics`, `PinholeIntrinsics`, and `Pinhole90Intrinsics` (see Utilities) classes are available in `data`, and provide the intrinsics values and matrix.\n\n```python\nK = CylindricalIntrinsics().K\n```\n\n\n## Structure\n\n### Data Location\nTo set the dataset download/loading location, you can use `data.set_download_location`, and `get_download_location` to get it.\nHowever, this is not advised.\nInstead, set the environment variable `CARLA_DOWNLOAD_LOCATION`.\nThe above methods are simply wrappers around it.\nIt is loaded using `pathlib.Path`, so use whatever path format your machine does.\n\n### Config\n\nEach simulation run is represented by a `config.Config` object, which contains the simulation parameters \n(city, rain, number of pedestrians, etc) necessary to get the location of the data files.\nIt provides utility methods and properties based on this, including getting data files for pinhole, cylindrical, \nspherical, and pose data, getting the local and remote locations, and downloading all of the data files.\n\n#### Loading\n\nA `Config` can be constructed manually or from the folder name (using `Config.from_folder_name`), \nbut will probably be loaded from `config.expand_wildcards`.\n`expand_wildcards` fills `None` parameters will all valid values.\nYou probably don\'t need to use it directly, as loader methods `load_df`, `load_csv`, and `load_text` methods are provided, as well as `all`.\n\n### Data Files\n\nData files can be gotten from `Config` objects, and represent a data file.\nThey have `download` methods that accept a `force` parameter, an `is_downloaded()` method, and a `data` property.\nNon-pose data files also have an `intrinsics` property to get the associated intrinsics (described later).\nThe `data` property opens and loads the file, returning a `DataFile`.\n`DataFile`s also work with Python\'s `with` blocks, and can be used like `with file as data:`.\nThis has the advantage of automatically closing the file. \nThe pinhole data file also exposes side methods `top`, `bottom`, etc. to get individual sides\'\ngroups within the pinhole data file, and a `__getitem__` operator that takes a `data.Side` to do the same.\n\n### Data\n\nThe data objects gotten from data files expose the ndarray data for the training run.\nPinhole data (gotten from pinhole data files) has properties to get the data for each side.\nPinhole side data (gotten from side data files), cylindrical data, and spherical data have `rgb` and `depth` properties\nthat return their respective data as `numpy` `ndarray`s.\nBoth are of shape `[batch, height, width, channels]`, where color images have 3 channels (RGB) and depth images have 1.\nDepth images are `uint16` and have the depth in decimeters, while color images are the standard `uint8`.\nPinhole images are 768x768, while cylindrical and spherical images are 2048x1024 (width x height).\n\n\nPose data objects have fields `absolute_pose`, `relative_pose`, and `start_relative_pose`.\nRelative pose is relative to the last pose value, while start relative pose is relative to the inital post of that simulation.\nPose data is shape `[batch, 6]`, where the 6 values are `[X, Y, Z, x, y, z]` where `[X, Y, Z]` is the position in meters, and `[x, y, z]` is the unit heading vector of the car.\n\n### Intrinsics\n`CylindricalIntrinsics`, `SphericalIntrinsics`, `PinholeIntrinsics`, and `Pinhole90Intrinsics` (see Utilities) are available in `data`, and provide the intrinsics values and matrix.\nEach object has `K`, `normalized_K`, `height`, `width`, `f_x`, `f_y`, `c_x`, `c_y`, and `fov` (degrees) fields.\n\n### Utilities\n\nA method `data.crop_pinhole_to_90` is provided to crop the 100 degree FOV pinhole images into 90 degree FOV images of the same size.\nIt uses `cv2.getRectSubPix` internally.\nThe intrinsics for these images are provided by `Pinhole90Intrinsics` using the same format as other intrinsics.\n',
    'author': 'Ryan Nett',
    'author_email': 'rnett@calpoly.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rnett/cpdd-dataset',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from enum import Enum
from pathlib import Path

from ..data import CylindricalDataFile, PinholeDataFile, SphericalDataFile
from ..data import get_download_location
from ..data._run_data import PoseDataFile


class Rain(Enum):
    Clear = 1
    Cloudy = 2
    Wet = 3
    WetCloudy = 4
    Soft = 5
    Mid = 6
    Hard = 7


class Weather:
    def __init__(self, rain: Rain, sunset: bool = False):
        self.rain = rain
        self.sunset = sunset


class City(Enum):
    Town01 = 1
    Town02 = 2
    Town03 = 3
    Town04 = 4
    Town05 = 5

    @property
    def map(self) -> str:
        return f"/Game/Carla/Maps/{self.name}"


class Config:

    def __init__(self, city: City, rain: Rain, sunset: bool, num_cars: int, num_peds: int, index: int):

        self.num_peds = num_peds
        self.num_cars = num_cars
        self.sunset = sunset
        self.rain = rain
        self.city = city
        self.index = index
        self.weather = Weather(rain, sunset)

    def __repr__(self):
        return f"Config(city={self.city.name}, rain={self.rain.name}, sunset={self.sunset}, num_cars={self.num_cars}, num_peds={self.num_peds}, index={self.index})"

    @property
    def folder_name(self) -> str:
        if self.sunset:
            time = 'sunset'
        else:
            time = 'noon'

        return f"{self.city.name.lower()}/" \
               f"{self.rain.name.lower()}/{time}/cars_{self.num_cars}_peds_" \
               f"{self.num_peds}_index_{self.index}"

    @property
    def download_location(self) -> Path:
        return get_download_location() / self.folder_name

    @property
    def remote_location(self) -> str:
        return "s3://cscdatasets/jventu09/cpdd_dataset/" + self.folder_name + "/"

    @property
    def cylindrical_data(self) -> CylindricalDataFile:
        return CylindricalDataFile(self)

    @property
    def spherical_data(self) -> SphericalDataFile:
        return SphericalDataFile(self)

    @property
    def pinhole_data(self) -> PinholeDataFile:
        return PinholeDataFile(self)

    @property
    def pose_data(self) -> PoseDataFile:
        return PoseDataFile(self)

    def download_all(self, force: bool = False):
        self.pose_data.download(force)
        self.cylindrical_data.download(force)
        self.spherical_data.download(force)
        self.pinhole_data.download(force)

    @staticmethod
    def from_folder_name(name: str):
        """
        name should be like "[/]{town}/{rain}/{time}/{folder_name}[/]"
        """

        parts = name.strip('/').split('/')

        town_name = parts[0]

        town = None

        for t in list(City):
            if t.name.lower() == town_name.lower():
                town = t
                break

        if town is None:
            raise ValueError(f"No town found for name {parts[0]}")

        rain_name = parts[1]

        rain = None

        for t in list(Rain):
            if t.name.lower() == rain_name.lower():
                rain = t
                break

        if rain is None:
            raise ValueError(f"No rain setting found for name {parts[1]}")

        sunset = parts[2] == "sunset"

        folder_name = parts[3].split('_')
        cars = int(folder_name[1])
        peds = int(folder_name[3])

        if len(folder_name) > 5:
            index = int(folder_name[5])
        else:
            index = 0

        return Config(town, rain, sunset, cars, peds, index)



from collections import deque
from pathlib import Path
from typing import List, Union

import pandas

from cpdd_dataset.config import City, Config, Rain


def cars_for_city(city: City):
    """
    The number of cars used for a city (necessary as city sizes vary).
    :param city:
    :return:
    """
    if city is City.Town01:
        return 30
    elif city is City.Town02:
        return 10
    elif city is City.Town03:
        return 40
    elif city is City.Town04:
        return 30
    elif city is City.Town05:
        return 40
    else:
        raise ValueError


def all() -> List[Config]:
    """
    :return: All existing Configs
    """
    return expand_wildcards()


def expand_wildcards(city=None, rain=None, sunset=None, num_cars=None, num_peds=None, index=None) -> List[Config]:
    """
    Expands any wildcard (None) values to existing values.  Passing all Nones returns all existing configs,
    the same as all().
    :param city:
    :param rain:
    :param sunset:
    :param num_cars:
    :param num_peds:
    :param index:
    :return:
    """
    done = []
    waiting = deque([(city, rain, sunset, num_cars, num_peds, index),] )

    while len(waiting) > 0:
        item = waiting.popleft()

        if item[0] is None:  # city
            for c in list(City):
                waiting.append((c, item[1], item[2], item[3], item[4], item[5]))
        elif item[1] is None: # rain
            # for r in list(Rain):
            #     waiting.append((item[0], r, item[2], item[3], item[4], item[5]))
            waiting.append((item[0], Rain.Clear, item[2], item[3], item[4], item[5]))
        elif item[2] is None: # sunset
            waiting.append((item[0], item[1], False, item[3], item[4], item[5]))
            # waiting.append((item[0], item[1], True, item[3], item[4], item[5]))
        elif item[3] is None: # cars
            waiting.append((item[0], item[1], item[2], cars_for_city(item[0]), item[4], item[5]))
        elif item[4] is None: # peds
            waiting.append((item[0], item[1], item[2], item[3], 200, item[5]))
        elif item[5] is None: # indices
            for i in range(2) if item[2] else range(5):
                waiting.append((item[0], item[1], item[2], item[3], item[4], i))
        else:
            done.append(item)

    return [Config(item[0], item[1], item[2], item[3], item[4], item[5]) for item in done]


def load_df(df: pandas.DataFrame) -> List[Config]:
    """
    Loads a list of configs from a pandas data frame.  Both column names and values are non case sensitive.

    |  Expects columns (non case sensitive):
    |  city: the name of a City
    |  rain: the name of a Rain
    |  time: sunset or noon
    |  num_cars: an int
    |  num_peds: an int
    |  index: an int

    | All values can be * to fill in all existing values.

    :param df: The data frame to load from
    :return: A list of Configs
    """

    # df.columns = map(str.lower, df.columns)
    configs = []

    for row in df.itertuples(index=False):
        num_cars = None if row.num_cars == "*" else int(row.num_cars)
        num_peds = None if row.num_peds == "*" else int(row.num_peds)
        index = None if row.index == "*" else int(row.index)
        city_name = None if row.city == "*" else row.city.lower()

        if city_name is not None:
            for c in list(City):
                if c.name.lower() == city_name:
                    city = c
                    break
            else:
                raise ValueError(f"{city_name} is not a valid city")
        else:
            city = None

        rain_name = None if row.rain == "*" else row.rain.lower()

        if rain_name is not None:
            for r in list(Rain):
                if r.name.lower() == rain_name:
                    rain = r
                    break
            else:
                raise ValueError(f"{rain_name} is not a valid rain")
        else:
            rain = None

        sunset_name = None if row.time == "*" else row.time.lower()
        if sunset_name is None:
            sunset = None
        elif sunset_name == "sunset":
            sunset = True
        elif sunset_name == "noon":
            sunset = False
        else:
            raise ValueError(f"{sunset_name} is not a valid time")

        configs.extend(expand_wildcards(city, rain, sunset, num_cars, num_peds, index))

    return configs


def load_txt(file: Union[Path, str]) -> List[Config]:
    """
    Loads a list of configs from a space-delimited text file.  Both column datas are non case sensitive.

    |  Expects columns (non case sensitive):
    |  city: the name of a City
    |  rain: the name of a Rain
    |  time: sunset or noon
    |  num_cars: an int
    |  num_peds: an int
    |  index: an int


    :param file: The file to read from
    :return: A list of Configs
    """

    if not isinstance(file, Path):
        file = Path(file)

    return load_df(pandas.read_csv(file, delim_whitespace=True))


def load_csv(file: Union[Path, str]) -> List[Config]:
    """
    Loads a list of configs from a comma delimited file.  Both column datas are non case sensitive.

    |  Expects columns (non case sensitive):
    |  city: the name of a City
    |  rain: the name of a Rain
    |  time: sunset or noon
    |  num_cars: an int
    |  num_peds: an int
    |  index: an int


    :param file: The file to read from
    :return: A list of Configs
    """

    if not isinstance(file, Path):
        file = Path(file)

    return load_df(pandas.read_csv(file))
